"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __asyncValues = (this && this.__asyncIterator) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator];
    return m ? m.call(o) : typeof __values === "function" ? __values(o) : o[Symbol.iterator]();
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r);  }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const LazyMap_1 = require("../Collections/LazyMap");
const Metadata_1 = require("../Metadata/Metadata");
class RedisLazyMap extends LazyMap_1.LazyMap {
    /**
     * Creates an instance of RedisLazyMap.
     * @param mapId Full map id in redis
     * @param manager Manager instance
     * @param [entityClass] If passed then map will be treated as entity map and will set/return entities
     * @param [cascadeInsert=false] True to automatically save entities
     */
    constructor(mapId, manager, entityClass, cascadeInsert = false) {
        super();
        this.manager = manager;
        this.mapId = mapId;
        this.entityClass = entityClass;
        this.cascadeInsert = cascadeInsert;
    }
    /**
     * Get map size
     *
     * @returns
     */
    size() {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.manager.connection.client.hlenAsync(this.mapId);
        });
    }
    /**
     * Set value
     *
     * @param key
     * @param value
     * @returns
     */
    set(key, value) {
        return __awaiter(this, void 0, void 0, function* () {
            const serializedKey = this.manager.serializeSimpleValue(key);
            if (serializedKey) {
                if (this.entityClass) {
                    const entityId = Metadata_1.getEntityFullId(value);
                    if (!entityId) {
                        throw new Error(`Unable to get entity id for key: ${key} and class ${this.entityClass.name}`);
                    }
                    if (this.cascadeInsert) {
                        // save entity
                        yield this.manager.save(value);
                    }
                    // set mapping in map itself
                    yield this.manager.connection.client.hsetAsync(this.mapId, serializedKey, entityId);
                }
                else {
                    const serializedVal = this.manager.serializeSimpleValue(value);
                    if (serializedVal) {
                        this.manager.connection.client.hsetAsync(this.mapId, serializedKey, serializedVal);
                    }
                }
            }
        });
    }
    /**
     * Delete value by key
     *
     * @param key
     * @param [deleteEntity=false] Also delete corresponding entity
     * @returns
     */
    delete(key, deleteEntity = false) {
        return __awaiter(this, void 0, void 0, function* () {
            let res;
            const serializedKey = this.manager.serializeSimpleValue(key);
            if (!serializedKey) {
                return false;
            }
            if (this.entityClass) {
                if (deleteEntity) {
                    const entityId = yield this.manager.connection.client.hgetAsync(this.mapId, serializedKey);
                    if (entityId) {
                        yield this.manager.removeById(this.entityClass, entityId);
                    }
                }
                res = yield this.manager.connection.client.hdelAsync(this.mapId, serializedKey);
            }
            else {
                res = yield this.manager.connection.client.hdelAsync(this.mapId, serializedKey);
            }
            return !!res;
        });
    }
    /**
     * Check if the map has key
     *
     * @param key
     * @returns
     */
    has(key) {
        return __awaiter(this, void 0, void 0, function* () {
            const serializedKey = this.manager.serializeSimpleValue(key);
            if (!serializedKey) {
                return false;
            }
            if (this.entityClass) {
                // check also if we have this entity itself
                const entityId = yield this.manager.connection.client.hgetAsync(this.mapId, serializedKey);
                if (!entityId) {
                    return false;
                }
                return !!(yield this.manager.connection.client.existsAsync(entityId));
            }
            else {
                // simple checking by checking key
                return !!(yield this.manager.connection.client.hexistsAsync(this.mapId, serializedKey));
            }
        });
    }
    /**
     * Get the value by given key
     *
     * @param key
     * @returns
     */
    get(key) {
        return __awaiter(this, void 0, void 0, function* () {
            const serializedKey = this.manager.serializeSimpleValue(key);
            if (!serializedKey) {
                return;
            }
            const val = yield this.manager.connection.client.hgetAsync(this.mapId, serializedKey);
            if (!val) {
                return;
            }
            // return simple value if not entity map
            if (!this.entityClass) {
                return this.manager.unserializeSimpleValue(val);
            }
            // Load entity otherwise
            return yield this.manager.load(this.entityClass, val);
        });
    }
    /**
     * Clear map
     *
     * @param [deleteEntities=false] Also delete entties
     * @returns
     */
    clear(deleteEntities = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.entityClass && deleteEntities) {
                const values = yield this.manager.connection.client.hvalsAsync(this.mapId);
                yield this.manager.removeById(this.entityClass, values);
            }
            // delete map
            yield this.manager.connection.client.delAsync(this.mapId);
        });
    }
    /**
     * Convert map to array
     *
     * @returns
     */
    toArray() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = [];
            try {
                for (var _a = __asyncValues(this), _b; _b = yield _a.next(), !_b.done;) {
                    const pair = yield _b.value;
                    res.push(pair);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_b && !_b.done && (_c = _a.return)) yield _c.call(_a);
                }
                finally { if (e_1) throw e_1.error; }
            }
            return res;
            var e_1, _c;
        });
    }
    /**
     * Iterate over map keys => values
     *
     * @returns
     */
    [Symbol.asyncIterator]() {
        return __asyncGenerator(this, arguments, function* _a() {
            let [cursor, results] = yield __await(this.manager.connection.client.hscanAsync(this.mapId, "0"));
            const getResults = (result) => __awaiter(this, void 0, void 0, function* () {
                const unserializedKeys = result.reduce((keys, current, index) => {
                    if (index % 2 === 0) {
                        keys.push(current);
                    }
                    return keys;
                }, []).map(val => this.manager.unserializeSimpleValue(val));
                let vals = result.reduce((vals, current, index) => {
                    if (index % 2) {
                        vals.push(current);
                    }
                    return vals;
                }, []);
                if (this.entityClass) {
                    // Loading will return only unique entities and since map can have different keys
                    // pointed to the same value the we need to preserve the original order
                    const entities = yield this.manager.load(this.entityClass, vals);
                    const entityMap = new Map();
                    for (const ent of entities) {
                        entityMap.set(Metadata_1.getEntityFullId(ent), ent);
                    }
                    vals = vals.map(val => entityMap.get(val));
                }
                else {
                    vals = vals.map(val => this.manager.unserializeSimpleValue(val));
                }
                return unserializedKeys.map((key, index) => [key, vals[index]]);
            });
            const res = yield __await(getResults(results));
            for (const keyVal of res) {
                yield keyVal;
            }
            while (cursor !== "0") {
                [cursor, results] = yield __await(this.manager.connection.client.hscanAsync(this.mapId, cursor));
                const res = yield __await(getResults(results));
                for (const keyVal of res) {
                    yield keyVal;
                }
            }
        });
    }
    /**
     * Iterate over map's keys
     *
     * @returns
     */
    keys() {
        return __asyncGenerator(this, arguments, function* keys_1() {
            let [cursor, results] = yield __await(this.manager.connection.client.hscanAsync(this.mapId, "0"));
            const getKeysFromResult = (result) => {
                return results.reduce((keys, current, index) => {
                    if (index % 2 === 0) {
                        keys.push(current);
                    }
                    return keys;
                }, []).map(val => this.manager.unserializeSimpleValue(val));
            };
            let keys = getKeysFromResult(results);
            for (const key of keys) {
                yield key;
            }
            while (cursor !== "0") {
                [cursor, results] = yield __await(this.manager.connection.client.hscanAsync(this.mapId, cursor));
                keys = getKeysFromResult(results);
                for (const key of keys) {
                    yield key;
                }
            }
        });
    }
    /**
     * Iterate over map's values
     *
     * @returns
     */
    values() {
        return __asyncGenerator(this, arguments, function* values_1() {
            try {
                for (var _a = __asyncValues(this), _b; _b = yield __await(_a.next()), !_b.done;) {
                    const [, val] = yield __await(_b.value);
                    yield val;
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (_b && !_b.done && (_c = _a.return)) yield __await(_c.call(_a));
                }
                finally { if (e_2) throw e_2.error; }
            }
            var e_2, _c;
        });
    }
}
exports.RedisLazyMap = RedisLazyMap;
//# sourceMappingURL=RedisLazyMap.js.map