"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __asyncValues = (this && this.__asyncIterator) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator];
    return m ? m.call(o) : typeof __values === "function" ? __values(o) : o[Symbol.iterator]();
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r);  }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const LazySet_1 = require("../Collections/LazySet");
const Metadata_1 = require("../Metadata/Metadata");
/**
 * Redis-backed lazy set
 *
 * @export
 * @class RedisLazySet
 * @template T
 */
class RedisLazySet extends LazySet_1.LazySet {
    /**
     * Creates an instance of RedisLazySet.
     * @param setId Full set id in redis
     * @param manager Manager instance
     * @param [entityClass] If passed then set will be treated as entity set and will return entities
     */
    constructor(setId, manager, entityClass, cascadeInsert = false) {
        super();
        this.manager = manager;
        this.setId = setId;
        this.entityClass = entityClass;
        this.cascadeInsert = cascadeInsert;
    }
    /**
     * Add value or entity to the set
     *
     * @param value
     * @returns
     */
    add(value) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.entityClass) {
                // Entity
                if (this.cascadeInsert) {
                    yield this.manager.save(value);
                }
                yield this.manager.connection.client.saddAsync(this.setId, Metadata_1.getEntityFullId(value));
            }
            else {
                const serializedVal = this.manager.serializeSimpleValue(value);
                if (serializedVal) {
                    yield this.manager.connection.client.saddAsync(this.setId, serializedVal);
                }
            }
        });
    }
    /**
     * Remove value or entity from the set.
     *
     * @param value
     * @param [deleteEntity=false]
     * @returns
     */
    delete(value, deleteEntity = false) {
        return __awaiter(this, void 0, void 0, function* () {
            let res;
            if (this.entityClass) {
                // Entity
                if (deleteEntity) {
                    yield this.manager.remove(value);
                }
                res = yield this.manager.connection.client.sremAsync(this.setId, Metadata_1.getEntityFullId(value));
            }
            else {
                const serializedVal = this.manager.serializeSimpleValue(value);
                if (serializedVal) {
                    res = yield this.manager.connection.client.sremAsync(this.setId, serializedVal);
                }
                else {
                    res = 0;
                }
            }
            return res > 0 ? true : false;
        });
    }
    /**
     * Determine if value or entity exists in the set
     *
     * @param value
     * @returns
     */
    has(value) {
        return __awaiter(this, void 0, void 0, function* () {
            const id = this.entityClass ? Metadata_1.getEntityFullId(value) : this.manager.serializeSimpleValue(value);
            if (typeof id === "undefined") {
                return false;
            }
            const res = yield this.manager.connection.client.sismemberAsync(this.setId, id);
            return !!res;
        });
    }
    /**
     * Get size of set
     *
     * @returns
     */
    size() {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.manager.connection.client.scardAsync(this.setId);
        });
    }
    /**
     * Clear set
     *
     * @param [deleteEntities=false] Also delete all entities
     * @returns
     */
    clear(deleteEntities = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.entityClass && deleteEntities) {
                const setVals = yield this.manager.connection.client.smembersAsync(this.setId);
                yield this.manager.removeById(this.entityClass, setVals);
            }
            yield this.manager.connection.client.delAsync(this.setId);
        });
    }
    /**
     * Convert set to array
     *
     * @returns
     */
    toArray() {
        return __awaiter(this, void 0, void 0, function* () {
            const results = [];
            try {
                for (var _a = __asyncValues(this.values()), _b; _b = yield _a.next(), !_b.done;) {
                    const v = yield _b.value;
                    results.push(v);
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_b && !_b.done && (_c = _a.return)) yield _c.call(_a);
                }
                finally { if (e_1) throw e_1.error; }
            }
            return results;
            var e_1, _c;
        });
    }
    /**
     * Iterate over values
     */
    values() {
        return __asyncGenerator(this, arguments, function* values_1() {
            let [cursor, results] = yield __await(this.manager.connection.client.sscanAsync(this.setId, "0"));
            // load entities
            if (this.entityClass && results.length > 0) {
                results = (yield __await(this.manager.load(this.entityClass, results)));
            }
            else {
                results = results.map(res => this.manager.unserializeSimpleValue(res));
            }
            for (const res of results) {
                yield res;
            }
            while (cursor !== "0") {
                [cursor, results] = yield __await(this.manager.connection.client.sscanAsync(this.setId, cursor));
                // load entities
                if (this.entityClass && results.length > 0) {
                    results = (yield __await(this.manager.load(this.entityClass, results)));
                }
                else {
                    results = results.map(res => this.manager.unserializeSimpleValue(res));
                }
                for (const res of results) {
                    yield res;
                }
            }
        });
    }
    /**
     * Iterator over values
     *
     * @returns
     */
    [Symbol.asyncIterator]() {
        return __asyncGenerator(this, arguments, function* _a() {
            try {
                for (var _a = __asyncValues(this.values()), _b; _b = yield __await(_a.next()), !_b.done;) {
                    const val = yield __await(_b.value);
                    yield val;
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (_b && !_b.done && (_c = _a.return)) yield __await(_c.call(_a));
                }
                finally { if (e_2) throw e_2.error; }
            }
            var e_2, _c;
        });
    }
}
exports.RedisLazySet = RedisLazySet;
//# sourceMappingURL=RedisLazySet.js.map