import { RedisManager } from "../Persistence/RedisManager";
import { EntitySubscriberInterface } from "../Subscriber/EntitySubscriberInterface";
import { PubSubSubscriberInterface } from "../Subscriber/PubSubSubscriberInterface";
import * as redis from "../utils/PromisedRedis";
export interface ConnectionOptions extends redis.ClientOpts {
    /**
     * Array of entity subscribers
     */
    entitySubscribers?: Array<{
        new (...args: any[]): EntitySubscriberInterface<any>;
    }>;
    /**
     * PubSub subscriber
     */
    pubSubSubscriber?: {
        new (...args: any[]): PubSubSubscriberInterface;
    };
}
export declare class Connection {
    /**
     * Manager instance
     */
    manager: RedisManager;
    /**
     * Connection status
     */
    isConnected: boolean;
    /**
     * Redis client instance
     */
    client: redis.RedisClient;
    /**
     * PubSub redis client
     *
     * @private
     */
    private pubsub;
    /**
     * Monitor redis client
     *
     * @private
     */
    private monitor?;
    /**
     * Options
     *
     * @private
     */
    private options;
    /**
     * Creates an instance of Connection.
     */
    constructor();
    /**
     * Init connection & connect
     *
     * @param options
     */
    connect(options: ConnectionOptions): Promise<void>;
    /**
     * Disconnect all connections
     *
     * @returns
     */
    disconnect(): Promise<void>;
    /**
     * Creates monitor connection and executes MONITOR command
     */
    startMonitoring(logger: (time: number, args: any[], reply: string) => void): Promise<void>;
    /**
     * Stops any monitoring
     *
     * @returns
     */
    stopMonitoring(): Promise<void>;
    /**
     * Subscribe for one or more chanels
     */
    subscribe(channel: string, ...channels: string[]): Promise<string>;
    /**
     * Stop listening for messages posted to the given channels.
     */
    unsubscribe(channel: string, ...channels: string[]): Promise<string>;
    /**
     * Stop listening for messages posted to channels matching the given patterns.
     */
    psubscribe(pattern: string, ...patterns: string[]): Promise<string>;
    /**
     * Stop listening for messages posted to channels matching the given patterns.
     */
    punsubscribe(pattern: string, ...channels: string[]): Promise<string>;
    /**
     * Remove all keys from the current database.
     */
    flushdb(): Promise<void>;
    /**
     * Execute multiple comamnds inside transaction
     *
     * @param executor
     * @param [atomic=false]
     * @returns
     */
    transaction(executor: (multi: redis.Commands<void>) => void, atomic?: boolean): Promise<any[]>;
    /**
     * Execute multiple commands in batch without transaction
     *
     * @param executor
     * @returns
     */
    batch(executor: (batch: redis.Commands<void>) => void): Promise<any[]>;
    /**
     * Init pubsub connection and listeners
     *
     * @private
     */
    private initPubSubListener();
    /**
     * Instantiate all entity subscribers
     *
     * @private
     * @param subscriberConstructors
     * @returns
     */
    private loadEntitySubscribers(subscriberConstructors?);
}
