"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Errors_1 = require("../Errors/Errors");
const RedisManager_1 = require("../Persistence/RedisManager");
const Container_1 = require("../utils/Container");
const redis = require("../utils/PromisedRedis");
class Connection {
    /**
     * Creates an instance of Connection.
     */
    constructor() {
        /**
         * Connection status
         */
        this.isConnected = false;
    }
    /**
     * Init connection & connect
     *
     * @param options
     */
    connect(options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.isConnected) {
                throw new Errors_1.AlreadyConnectedError();
            }
            this.options = options;
            this.client = redis.createClient(options);
            this.pubsub = redis.createClient(options);
            try {
                this.initPubSubListener();
                const subscribers = this.loadEntitySubscribers(this.options.entitySubscribers);
                this.manager = new RedisManager_1.RedisManager(this, subscribers);
                this.isConnected = true;
            }
            catch (e) {
                yield this.disconnect();
                throw e;
            }
        });
    }
    /**
     * Disconnect all connections
     *
     * @returns
     */
    disconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            // Stop monitor connection if exist
            if (this.monitor) {
                yield this.monitor.quitAsync();
            }
            // Stop pubsub connection
            if (this.pubsub) {
                yield this.pubsub.quitAsync();
            }
            // Stop client connection
            if (this.client) {
                yield this.client.quitAsync();
            }
            this.isConnected = false;
        });
    }
    /**
     * Creates monitor connection and executes MONITOR command
     */
    startMonitoring(logger) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.monitor) {
                return;
            }
            this.monitor = redis.createClient(this.options);
            this.monitor.on("monitor", logger);
            yield this.monitor.monitorAsync();
        });
    }
    /**
     * Stops any monitoring
     *
     * @returns
     */
    stopMonitoring() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.monitor) {
                yield this.monitor.quitAsync();
                this.monitor = undefined;
            }
        });
    }
    /**
     * Subscribe for one or more chanels
     */
    subscribe(channel, ...channels) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.pubsub.subscribeAsync(channel, ...channels);
        });
    }
    /**
     * Stop listening for messages posted to the given channels.
     */
    unsubscribe(channel, ...channels) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.pubsub.unsubscribeAsync(channel, ...channels);
        });
    }
    /**
     * Stop listening for messages posted to channels matching the given patterns.
     */
    psubscribe(pattern, ...patterns) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.pubsub.psubscribeAsync(pattern, ...patterns);
        });
    }
    /**
     * Stop listening for messages posted to channels matching the given patterns.
     */
    punsubscribe(pattern, ...channels) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.pubsub.punsubscribeAsync(pattern, ...channels);
        });
    }
    /**
     * Remove all keys from the current database.
     */
    flushdb() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.client.flushdbAsync();
        });
    }
    /**
     * Execute multiple comamnds inside transaction
     *
     * @param executor
     * @param [atomic=false]
     * @returns
     */
    transaction(executor, atomic = false) {
        return __awaiter(this, void 0, void 0, function* () {
            const multi = this.client.multi();
            executor(multi);
            return new Promise((resolve, reject) => {
                const func = atomic ? multi.exec_atomic : multi.exec;
                func.call(multi, (err, result) => {
                    err ? reject(err) : resolve(result);
                });
            });
        });
    }
    /**
     * Execute multiple commands in batch without transaction
     *
     * @param executor
     * @returns
     */
    batch(executor) {
        return __awaiter(this, void 0, void 0, function* () {
            const batch = this.client.batch();
            executor(batch);
            return new Promise((resolve, reject) => {
                batch.exec((err, result) => {
                    err ? reject(err) : resolve(result);
                });
            });
        });
    }
    /**
     * Init pubsub connection and listeners
     *
     * @private
     */
    initPubSubListener() {
        if (typeof this.options.pubSubSubscriber === "function") {
            const listener = Container_1.getFromContainer(this.options.pubSubSubscriber);
            if (listener.onMessage) {
                this.pubsub.on("message", listener.onMessage.bind(listener));
            }
            if (listener.onMessageBuffer) {
                this.pubsub.on("message_buffer", listener.onMessageBuffer.bind(listener));
            }
            if (listener.onPMessage) {
                this.pubsub.on("pmessage", listener.onPMessage.bind(listener));
            }
            if (listener.onPMessageBuffer) {
                this.pubsub.on("pmessage_buffer", listener.onPMessageBuffer.bind(listener));
            }
            if (listener.onSubscribe) {
                this.pubsub.on("subscribe", listener.onSubscribe.bind(listener));
            }
            if (listener.onPSubscribe) {
                this.pubsub.on("psubscribe", listener.onPSubscribe.bind(listener));
            }
            if (listener.onUnsubscribe) {
                this.pubsub.on("unsubscribe", listener.onUnsubscribe.bind(listener));
            }
            if (listener.onPUnsubscribe) {
                this.pubsub.on("punsubscribe", listener.onPUnsubscribe.bind(listener));
            }
        }
    }
    /**
     * Instantiate all entity subscribers
     *
     * @private
     * @param subscriberConstructors
     * @returns
     */
    loadEntitySubscribers(subscriberConstructors) {
        const subscribers = [];
        if (subscriberConstructors) {
            for (const cl of subscriberConstructors) {
                subscribers.push(Container_1.getFromContainer(cl));
            }
        }
        return subscribers;
    }
}
exports.Connection = Connection;
//# sourceMappingURL=Connection.js.map