/**
 * Metadata key to mark class to be stored as redis hash
 */
export declare const REDIS_ENTITY: symbol;
/**
 * Metadata key to store redis properties metadatas
 */
export declare const REDIS_PROPERTIES: symbol;
/**
 * Metadata key to keep serialized initial value of property from redis. For collections this will be link to the corresponding hashmap/set
 */
export declare const REDIS_VALUE: symbol;
/**
 * Metadata key to keep the initial state of redis collections (map/set)
 */
export declare const REDIS_COLLECTION_VALUE: symbol;
export declare type PropertyMetadata = SimplePropertyMetadata | RelationPropertyMetadata;
/**
 * Hash simple property metadata
 *
 * @export
 * @interface SimplePropertyMetadata
 */
export interface SimplePropertyMetadata {
    /**
     * Property name in class
     */
    propertyName: string;
    /**
     * Property name stored in redis
     */
    propertyRedisName: string;
    /**
     * Property type
     */
    propertyType: any;
    /**
     * True if property must be identify column
     */
    isIdentifyColumn: boolean;
    /**
     * Relation flag
     */
    isRelation: false;
}
/**
 * Relation options
 *
 * @export
 * @interface RelationOptions
 */
export interface RelationOptions {
    /**
     * Property name in redis
     */
    propertyName?: string;
    /**
     * True to automatically create new hash for relation
     */
    cascadeInsert?: boolean;
    /**
     * True to automatically update corresponding relation hash
     */
    cascadeUpdate?: boolean;
}
/**
 * Hash relation property metadata
 *
 * @export
 * @interface RelationPropertyMetadata
 */
export interface RelationPropertyMetadata {
    /**
     * Property name in class
     */
    propertyName: string;
    /**
     * Property name stored in redis
     */
    propertyRedisName: string;
    /**
     * Property type.
     * For set property relation it will be Set itself
     * For single property relation it will be equal to relationType
     */
    propertyType: any;
    /**
     * Relation flag
     */
    isRelation: true;
    /**
     * Relation type. Must be type of related entity
     */
    relationType: any;
    /**
     * Relation options
     */
    relationOptions: RelationOptions;
}
/**
 * Return true if given object is redis entity like
 *
 * @export
 * @param entity
 * @returns
 */
export declare function isRedisEntity(entity: object): boolean;
/**
 * Return entity name
 *
 * @export
 * @param entityOrEntityClass
 * @returns
 */
export declare function getEntityName(entityOrEntityClass: object | Function): string | undefined;
/**
 * Return entity id
 *
 * @export
 * @param entity
 * @returns
 */
export declare function getEntityId(entity: {
    [key: string]: any;
}): string | number | undefined;
/**
 * Returns redis properties metadata
 *
 * @export
 * @param entityOrEntityClass
 * @returns
 */
export declare function getEntityProperties(entityOrEntityClass: object | Function): PropertyMetadata[] | undefined;
/**
 * Return full hash id
 *
 * @export
 * @param entityOrEntityClass
 * @param [hashId]
 * @returns
 */
export declare function getEntityFullId(entityOrEntityClass: object | Function, hashId?: string | number): string | undefined;
