import { PropertyMetadata } from "../Metadata/Metadata";
/**
 * Hash key add/remove operation. This applies for creating new hash and for modifying existing hash
 *
 * @export
 * @interface HashPersistenceOperator
 */
export interface HashPersistenceOperator {
    /**
     * Hash id
     */
    hashId: string;
    /**
     * Keys to change/add. Existing keys will be owerwritten, news keys added
     */
    changeKeys: {
        [key: string]: string;
    };
    /**
     * Array of keys to delete in hash
     */
    deleteKeys: string[];
}
/**
 * Set add/remove operation. This applies for new sets and existings sets
 *
 * @export
 * @interface SetPersistenceOperator
 */
export interface SetPersistenceOperator {
    /**
     * Set id
     */
    setName: string;
    /**
     * Values in set to remove
     */
    removeValues: string[];
    /**
     * Values in set to add
     */
    addValues: string[];
}
/**
 * Persistence operation for given entity and it's relations
 *
 * @export
 * @interface PersistenceOperation
 */
export interface PersistenceOperation {
    /**
     * Array of hashes to create/modify
     */
    modifyHashes: HashPersistenceOperator[];
    /**
     * Array of hash ids to delete
     */
    deleteHashes: string[];
    /**
     * Array of sets to create/modify
     */
    modifySets: SetPersistenceOperator[];
    /**
     * Array of sets to delete
     */
    deletesSets: string[];
}
export interface LoadOperation {
    /**
     * Full entity id to load
     */
    entityId: string;
    /**
     * Additional Sets to load
     */
    sets: string[];
    /**
     * Additional Hashes to load
     */
    hashes: string[];
    /**
     * Relation mapping
     */
    relationMappings: Array<{
        /**
         * Owner hash id
         */
        ownerId: string;
        /**
         * Relation class
         */
        relationClass: Function;
        /**
         * Mapping type for relation value
         */
        type: "key" | "map" | "set";
        /**
         * Mapping id. for key it's the hash key, for map/sets it's the map/set full id
         */
        id: string;
    }>;
}
export interface HydrationData {
    id: string;
    /**
     * Class constructor if the data is entity properties
     */
    entityClass?: Function;
    /**
     * Raw data
     */
    redisData: {
        [key: string]: string;
    } | string[] | null;
}
export interface EntityWithId {
    entityId: string;
    entity: object;
}
/**
 * Object Operator
 *
 * @export
 * @class Operator
 */
export declare class Operator {
    /**
     * Prepare to save/update entity and get necessary operations
     *
     * @param entity
     * @returns
     */
    getSaveOperation(entity: {
        [key: string]: any;
    }, processedEntities?: EntityWithId[]): Promise<PersistenceOperation>;
    /**
     * Get delete operation for given loaded hash
     *
     * @param entityOrEntityClass
     * @param [entityId]
     * @returns
     */
    getDeleteOperation(entityOrEntityClass: {
        [key: string]: any;
    } | Function, entityId?: string | number): PersistenceOperation;
    /**
     * Get load operation for given hash class with given id
     *
     * @param id
     * @param hashClass
     * @param skipRelations
     * @returns
     */
    getLoadOperation(id: string | number, hashClass: Function, skipRelations?: string[]): LoadOperation | undefined;
    /**
     * Recursively process hash and set/update metadata
     *
     * @param hashObject
     * @param processedHashes
     */
    updateMetadataInHash(hashObject: {
        [key: string]: any;
    }, processedHashes?: object[]): void;
    /**
     * Recursively process given hash and reset metadata
     *
     * @param entityObj
     */
    resetMetadataInEntityObject(entityObj: {
        [key: string]: any;
    }): void;
    /**
     * Hydrate redis data
     *
     * @param hydrationData
     * @returns
     */
    hydrateData(hydrationData: Map<string, HydrationData>): any[];
    /**
     * Serialize value for persiting
     *
     * @private
     * @param value
     * @returns
     */
    serializeValue(value: any): string | undefined;
    /**
     * Unserialize value
     *
     * @private
     * @param value
     * @returns
     */
    unserializeValue(value: string): string | number | boolean | null | object | Date | undefined;
    /**
     * Get id for collection for property
     *
     * @param entityId
     * @param propertyMetadata
     * @returns
     */
    getCollectionId(entityId: string, propertyMetadata: PropertyMetadata): string | undefined;
    /**
     * Check metadata correctness
     *
     * @private
     * @param entity
     */
    private checkMetadata(entity);
    /**
     * Get entity name
     *
     * @private
     * @param entity
     * @returns
     */
    private getFullIdForEntityObject(entity);
    /**
     * Get full id for given hash class and id
     *
     * @private
     * @param entityClass
     * @param id
     * @returns
     */
    private getFullIdForEntityClass(entityClass, id);
    /**
     * Make sure that relations are valid and doesn't have same ids but different object links
     * Same id and same object link (i.e rel === rel) is allowed
     *
     * @private
     * @param relations
     * @param parentEntity
     */
    private ensureValidRelations(relations, parentEntity);
    /**
     * Prepare relation value persisting. This will be just entity name
     *
     * @private
     * @param value
     * @param metadata
     * @returns
     */
    private prepareRelationValue(value, metadata);
}
