"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const LazyMap_1 = require("../Collections/LazyMap");
const LazySet_1 = require("../Collections/LazySet");
const RedisLazyMap_1 = require("../Collections/RedisLazyMap");
const RedisLazySet_1 = require("../Collections/RedisLazySet");
const Errors_1 = require("../Errors/Errors");
const Metadata_1 = require("../Metadata/Metadata");
const hasPrototypeOf_1 = require("../utils/hasPrototypeOf");
/**
 * Object Operator
 *
 * @export
 * @class Operator
 */
class Operator {
    /**
     * Prepare to save/update entity and get necessary operations
     *
     * @param entity
     * @returns
     */
    getSaveOperation(entity, processedEntities = []) {
        return __awaiter(this, void 0, void 0, function* () {
            this.checkMetadata(entity);
            const metadatas = Reflect.getMetadata(Metadata_1.REDIS_PROPERTIES, entity.constructor);
            const fullHashId = this.getFullIdForEntityObject(entity);
            const operation = {
                deleteHashes: [],
                deletesSets: [],
                modifyHashes: [],
                modifySets: []
            };
            // Circular objects checking
            // Must check for circular references and prevent cascade delete/insert/update
            // Must check relation id and object equality to prevent different objects to be saved with same id
            const processedThisEntity = processedEntities.find(ent => ent.entityId === fullHashId);
            if (processedThisEntity) {
                if (processedThisEntity.entity !== entity) {
                    // we have proceed entity with same id but objects are differs. This is error 
                    throw new Errors_1.DuplicateIdsInEntityError(entity, fullHashId);
                }
                else {
                    // Return empty operation since we already proceed it
                    return operation;
                }
            }
            processedEntities.push({ entity: entity, entityId: fullHashId });
            const hashPropOperation = {
                hashId: fullHashId,
                // hashName: hashName,
                changeKeys: {},
                deleteKeys: []
            };
            operation.modifyHashes.push(hashPropOperation);
            for (const propMetadata of metadatas) {
                // Get initial redis value. Will be undefined if it's new entity. May contain "null" as string
                const initialPropertyRedisValue = Reflect.getMetadata(Metadata_1.REDIS_VALUE, entity, propMetadata.propertyName);
                const value = entity[propMetadata.propertyName];
                const valueType = propMetadata.propertyType;
                // special case for Map/Set
                if (hasPrototypeOf_1.hasPrototypeOf(valueType, Set) || hasPrototypeOf_1.hasPrototypeOf(valueType, Map) ||
                    hasPrototypeOf_1.hasPrototypeOf(valueType, LazySet_1.LazySet) || hasPrototypeOf_1.hasPrototypeOf(valueType, LazyMap_1.LazyMap)) {
                    const redisInitialCollectionValue = Reflect.getMetadata(Metadata_1.REDIS_COLLECTION_VALUE, entity, propMetadata.propertyName);
                    // LazyMaps and LazySets are not instances of Set/Maps so they won't be deleted
                    const deleted = !value || ((value instanceof Set || value instanceof Map) && value.size === 0);
                    const collectionId = this.getCollectionId(fullHashId, propMetadata);
                    if (deleted) {
                        // Deleted or wasn't set
                        // 1. had val -> null. Delete map/set and set prop to "null"
                        // 2. had val -> undefined. Delete map/set and delete prop
                        // 3. had null -> null. Nop
                        // 4. had null -> undefined. Delete prop
                        // 5. had undefined -> undefined. Nop
                        // 6. had undefined -> null. Add "null" prop
                        // 7. had val -> become empty set or map
                        if (initialPropertyRedisValue && initialPropertyRedisValue !== "null" && redisInitialCollectionValue) {
                            // 1, 2, 7
                            if (hasPrototypeOf_1.hasPrototypeOf(valueType, Set)) {
                                operation.deletesSets.push(collectionId);
                            }
                            else if (hasPrototypeOf_1.hasPrototypeOf(valueType, Map)) {
                                operation.deleteHashes.push(collectionId);
                            }
                            value === null
                                ? hashPropOperation.changeKeys[propMetadata.propertyRedisName] = this.serializeValue(null)
                                : hashPropOperation.deleteKeys.push(propMetadata.propertyRedisName);
                        }
                        else if (initialPropertyRedisValue === "null") {
                            // 3, 4
                            if (typeof value === "undefined") {
                                // 4
                                hashPropOperation.deleteKeys.push(propMetadata.propertyRedisName);
                            }
                        }
                        else if (!initialPropertyRedisValue) {
                            // 5, 6
                            if (value === null) {
                                // 6
                                hashPropOperation.changeKeys[propMetadata.propertyRedisName] = this.serializeValue(null);
                            }
                        }
                        // 
                    }
                    else if ((value instanceof Set || value instanceof Map) && value.size > 0) {
                        // deleted keys
                        const deleted = new Set();
                        // For non relations sets key and value will be same
                        const added = new Map();
                        const changed = new Map();
                        if (propMetadata.isRelation && (propMetadata.relationOptions.cascadeInsert || propMetadata.relationOptions.cascadeUpdate)) {
                            this.ensureValidRelations([...value.values()], entity);
                        }
                        // calculate key/values differences
                        if (value instanceof Set) {
                            // convert value set to map of values/entity names -> entity if needed, since original redis collection will only have them
                            const preparedValueMap = new Map();
                            for (const val of value.values()) {
                                if (propMetadata.isRelation) {
                                    const relationId = this.prepareRelationValue(val, propMetadata);
                                    preparedValueMap.set(relationId, val);
                                }
                                else {
                                    const preparedVal = this.serializeValue(val);
                                    if (preparedVal) {
                                        preparedValueMap.set(preparedVal, preparedVal);
                                    }
                                }
                            }
                            // Calculate added/changed values over original collection. cascadeUpdate for relations will always mark collection as changed
                            if (redisInitialCollectionValue && Array.isArray(redisInitialCollectionValue)) {
                                for (const key of preparedValueMap.keys()) {
                                    if (!redisInitialCollectionValue.includes(key)) {
                                        added.set(key, preparedValueMap.get(key));
                                    }
                                    else if (propMetadata.isRelation && propMetadata.relationOptions.cascadeUpdate) {
                                        changed.set(key, preparedValueMap.get(key));
                                    }
                                }
                                // calculate deleted values over original collection
                                for (const key of redisInitialCollectionValue) {
                                    if (!preparedValueMap.has(key)) {
                                        deleted.add(key);
                                    }
                                }
                            }
                            else {
                                // new set
                                for (const key of preparedValueMap.keys()) {
                                    added.set(key, preparedValueMap.get(key));
                                }
                            }
                        }
                        else if (value instanceof Map) {
                            if (redisInitialCollectionValue && !Array.isArray(redisInitialCollectionValue)) {
                                // Calculate added/changed values over original collection. cascadeUpdate for relations will always mark collection as changed
                                for (const key of value.keys()) {
                                    const preparedKey = this.serializeValue(key);
                                    if (!preparedKey) {
                                        continue;
                                    }
                                    // added new key to map
                                    if (typeof redisInitialCollectionValue[preparedKey] === "undefined") {
                                        added.set(preparedKey, value.get(key));
                                    }
                                    else {
                                        const preparedValue = propMetadata.isRelation
                                            ? this.prepareRelationValue(value.get(key), propMetadata)
                                            : this.serializeValue(value.get(key));
                                        if ((redisInitialCollectionValue[preparedKey] !== preparedValue ||
                                            (propMetadata.isRelation && propMetadata.relationOptions.cascadeUpdate)) // cascade update will always force
                                        ) {
                                            changed.set(preparedKey, value.get(key));
                                        }
                                    }
                                }
                                // key is preparedKey
                                for (const key of Object.keys(redisInitialCollectionValue)) {
                                    const convertedKey = this.unserializeValue(key);
                                    if (typeof convertedKey !== "undefined" && !value.has(convertedKey)) {
                                        deleted.add(key);
                                    }
                                }
                            }
                            else {
                                // new map
                                for (const key of value.keys()) {
                                    const preparedKey = this.serializeValue(key);
                                    if (typeof preparedKey === "undefined") {
                                        continue;
                                    }
                                    added.set(preparedKey, value.get(key));
                                }
                            }
                        }
                        if (initialPropertyRedisValue !== collectionId) {
                            // Add link to set/map
                            hashPropOperation.changeKeys[propMetadata.propertyRedisName] = collectionId;
                        }
                        // calculate differences from original collections
                        if (value instanceof Set) {
                            operation.modifySets.push({
                                setName: collectionId,
                                addValues: [...added.keys()],
                                removeValues: [...deleted.keys()]
                            });
                        }
                        else if (value instanceof Map) {
                            operation.modifyHashes.push({
                                hashId: collectionId,
                                changeKeys: [...added.keys(), ...changed.keys()].reduce((obj, key) => {
                                    const val = propMetadata.isRelation
                                        ? this.getFullIdForEntityObject(added.has(key) ? added.get(key) : changed.get(key))
                                        : this.serializeValue(added.has(key) ? added.get(key) : changed.get(key));
                                    if (val) {
                                        obj[key] = val;
                                    }
                                    return obj;
                                }, {}),
                                deleteKeys: [...deleted.values()]
                            });
                        }
                        // Update/Insert relations if needed
                        if (propMetadata.isRelation && (propMetadata.relationOptions.cascadeInsert || propMetadata.relationOptions.cascadeUpdate)) {
                            const relationOperations = yield Promise.all([
                                ...propMetadata.relationOptions.cascadeInsert ? added.values() : [],
                                ...propMetadata.relationOptions.cascadeUpdate ? changed.values() : []
                            ].map(entity => this.getSaveOperation(entity, processedEntities)));
                            for (const relOp of relationOperations) {
                                operation.deleteHashes.push(...relOp.deleteHashes);
                                operation.deletesSets.push(...relOp.deletesSets);
                                operation.modifyHashes.push(...relOp.modifyHashes);
                                operation.modifySets.push(...relOp.modifySets);
                            }
                        }
                    }
                    else if ((value instanceof LazyMap_1.LazyMap || value instanceof LazySet_1.LazySet)
                        && !(value instanceof RedisLazyMap_1.RedisLazyMap || value instanceof RedisLazySet_1.RedisLazySet) // don't process wrapped maps
                    ) {
                        if ((yield value.size()) > 0) {
                            const collArray = yield value.toArray();
                            if (value instanceof LazySet_1.LazySet) {
                                const serializedValues = collArray
                                    .map(val => propMetadata.isRelation ? this.getFullIdForEntityObject(val) : this.serializeValue(val))
                                    .filter(val => typeof val !== "undefined");
                                operation.modifySets.push({
                                    setName: collectionId,
                                    addValues: serializedValues,
                                    removeValues: []
                                });
                            }
                            else if (value instanceof LazyMap_1.LazyMap) {
                                const serializedParis = collArray
                                    .map(pair => {
                                    const serializedKey = this.serializeValue(pair[0]);
                                    const serializedValue = propMetadata.isRelation ? this.getFullIdForEntityObject(pair[1]) : this.serializeValue(pair[1]);
                                    if (typeof serializedKey !== "undefined" && typeof serializedValue !== "undefined") {
                                        return [serializedKey, serializedValue];
                                    }
                                    else {
                                        return undefined;
                                    }
                                }).filter(pair => typeof pair !== "undefined");
                                operation.modifyHashes.push({
                                    hashId: collectionId,
                                    deleteKeys: [],
                                    changeKeys: serializedParis.reduce((obj, [key, val]) => {
                                        obj[key] = val;
                                        return obj;
                                    }, {})
                                });
                            }
                            if (propMetadata.isRelation && propMetadata.relationOptions.cascadeInsert) {
                                const relationOperations = yield Promise.all(collArray
                                    .filter(val => Array.isArray(val) ? typeof val[1] !== "undefined" : typeof val !== "undefined")
                                    .map(val => Array.isArray(val) ? this.getSaveOperation(val[1], processedEntities) : this.getSaveOperation(val, processedEntities)));
                                for (const relOp of relationOperations) {
                                    operation.deleteHashes.push(...relOp.deleteHashes);
                                    operation.deletesSets.push(...relOp.deletesSets);
                                    operation.modifyHashes.push(...relOp.modifyHashes);
                                    operation.modifySets.push(...relOp.modifySets);
                                }
                            }
                        }
                    }
                }
                else {
                    // Ordinary hash value or single relation
                    const preparedValue = propMetadata.isRelation
                        ? this.prepareRelationValue(value, propMetadata)
                        : this.serializeValue(value);
                    if (typeof preparedValue === "undefined" && initialPropertyRedisValue) {
                        // deletion
                        // delete hash key
                        hashPropOperation.deleteKeys.push(propMetadata.propertyRedisName);
                    }
                    else if (typeof preparedValue !== "undefined") {
                        // changed, addition
                        // Update hash key in original entity if needed
                        if (preparedValue !== initialPropertyRedisValue) {
                            hashPropOperation.changeKeys[propMetadata.propertyRedisName] = preparedValue;
                        }
                        if (propMetadata.isRelation) {
                            // Cases
                            // 1. had relation -> set null, cascade delete true -> delete relation
                            // 2. had relation -> set null, cascade delete false -> Nop
                            // 3. had relation -> change relation object to new id, cascade insert true -> add new relation
                            // 4. had relation -> change relation object to new id, cascade insert false -> Nop
                            // 5. had relation -> change relation object props, cascade update true -> save relation
                            // 6. had relation -> change relation object props, cascade update false -> Nop
                            // 7. didn't have relation -> add relation, cascade insert true -> add new relation
                            // 8. didn't have relation -> add relation, cascade insert false -> Nop
                            // 9. had relation -> change relation object to new id, cascade delete true -> delete previous relation
                            // 1, 2, 9 are invalid now
                            // if (preparedValue === "null" && preparedValue !== initialPropertyRedisValue && propMetadata.relationOptions.cascadeDelete) {
                            //     // Case 1
                            //     // Delete relation
                            //     const originalRelationValue: any = Reflect.getMetadata(REDIS_RELATION_VALUE, entity, propMetadata.propertyName);
                            //     if (originalRelationValue) {
                            //         const relationCascadeDeleteOp = this.getDeleteOperation(originalRelationValue, processedEntities);
                            //         operation.deleteHashes.push(...relationCascadeDeleteOp.deleteHashes);
                            //         operation.deletesSets.push(...relationCascadeDeleteOp.deletesSets);
                            //     } else if (initialPropertyRedisValue && initialPropertyRedisValue !== "null") {
                            //         operation.deleteHashes.push(initialPropertyRedisValue);
                            //     }
                            // } else
                            if ((preparedValue !== initialPropertyRedisValue && propMetadata.relationOptions.cascadeInsert) ||
                                (preparedValue === initialPropertyRedisValue && propMetadata.relationOptions.cascadeUpdate) // case 5
                            ) {
                                if (propMetadata.relationOptions.cascadeInsert || propMetadata.relationOptions.cascadeUpdate) {
                                    const relationOperation = yield this.getSaveOperation(value, processedEntities);
                                    operation.deleteHashes.push(...relationOperation.deleteHashes);
                                    operation.deletesSets.push(...relationOperation.deletesSets);
                                    operation.modifyHashes.push(...relationOperation.modifyHashes);
                                    operation.modifySets.push(...relationOperation.modifySets);
                                }
                                // case 9
                                // if (typeof initialPropertyRedisValue !== "undefined" &&
                                //     preparedValue !== initialPropertyRedisValue &&
                                //     initialPropertyRedisValue !== "null" &&
                                //     propMetadata.relationOptions.cascadeDelete
                                // ) {
                                //     const originalRelationValue: any = Reflect.getMetadata(REDIS_RELATION_VALUE, entity, propMetadata.propertyName);
                                //     if (originalRelationValue) {
                                //         const relationCascadeDeleteOp = this.getDeleteOperation(originalRelationValue, processedEntities);
                                //         operation.deleteHashes.push(...relationCascadeDeleteOp.deleteHashes);
                                //         operation.deletesSets.push(...relationCascadeDeleteOp.deletesSets);
                                //     } else {
                                //         operation.deleteHashes.push(initialPropertyRedisValue);
                                //     }
                                // }
                            }
                        }
                    }
                }
            }
            return operation;
        });
    }
    /**
     * Get delete operation for given loaded hash
     *
     * @param entityOrEntityClass
     * @param [entityId]
     * @returns
     */
    getDeleteOperation(entityOrEntityClass, entityId) {
        const operation = {
            deleteHashes: [],
            deletesSets: [],
            modifyHashes: [],
            modifySets: []
        };
        if (!entityOrEntityClass) {
            return operation;
        }
        this.checkMetadata(entityOrEntityClass);
        const metadatas = Metadata_1.getEntityProperties(entityOrEntityClass);
        if (!metadatas) {
            throw new Errors_1.MetadataError(typeof entityOrEntityClass === "function" ? entityOrEntityClass : entityOrEntityClass.constructor, "Not an entity");
        }
        // const fullHashId = this.getFullIdForEntityObject(entityOrEntityClass);
        const fullHashId = typeof entityId === "string" && entityId.startsWith("e:")
            ? entityId
            : Metadata_1.getEntityFullId(entityOrEntityClass, entityId);
        if (!fullHashId) {
            throw new Errors_1.MetadataError(typeof entityOrEntityClass === "function" ? entityOrEntityClass : entityOrEntityClass.constructor, "Unable to get id from entity");
        }
        // Circular objects checking
        // not needed since we won't cascade
        // Must check for circular references and prevent delete/insert/update
        // Must check relation id and object equality to prevent different objects to be saved with same id
        // const processedThisEntity = processedEntities.find(ent => ent.entityId === fullHashId);
        // if (processedThisEntity) {
        //     if (processedThisEntity.entity !== entity) {
        //         // we have proceed entity with same id but objects are differs. This is error 
        //         throw new DuplicateIdsInEntityError(entity, fullHashId);
        //     } else {
        //         // Return empty operation since we already proceed it
        //         return operation;
        //     }
        // }
        // processedEntities.push({ entity: entity, entityId: fullHashId });
        operation.deleteHashes.push(fullHashId);
        for (const propMetadata of metadatas) {
            const valueType = propMetadata.propertyType;
            // const initialRedisValue: string | undefined | "null" = Reflect.getMetadata(REDIS_VALUE, entityOrEntityClass, propMetadata.propertyName);
            // always try to delete maps/sets, there won't be error if there are not exist
            if (hasPrototypeOf_1.hasPrototypeOf(valueType, Set) || hasPrototypeOf_1.hasPrototypeOf(valueType, LazySet_1.LazySet)) {
                // if (initialRedisValue && initialRedisValue !== "null") {
                operation.deletesSets.push(this.getCollectionId(fullHashId, propMetadata));
                // }
            }
            else if (hasPrototypeOf_1.hasPrototypeOf(valueType, Map) || hasPrototypeOf_1.hasPrototypeOf(valueType, LazyMap_1.LazyMap)) {
                // if (initialRedisValue && initialRedisValue !== "null") {
                operation.deleteHashes.push(this.getCollectionId(fullHashId, propMetadata));
                // }
            }
            // else if (propMetadata.isRelation && propMetadata.relationOptions.cascadeDelete && value) {
            //     const relationOperation = this.getDeleteOperation(value, processedEntities);
            //     operation.deleteHashes.push(...relationOperation.deleteHashes);
            //     operation.deletesSets.push(...relationOperation.deletesSets);
            // }
        }
        return operation;
    }
    /**
     * Get load operation for given hash class with given id
     *
     * @param id
     * @param hashClass
     * @param skipRelations
     * @returns
     */
    getLoadOperation(id, hashClass, skipRelations = []) {
        this.checkMetadata(hashClass);
        const fullHashId = typeof id === "string" && id.startsWith("e:") ? id : this.getFullIdForEntityClass(hashClass, id);
        const operation = {
            entityId: fullHashId,
            hashes: [],
            sets: [],
            relationMappings: []
        };
        if (id === "null") {
            return undefined;
        }
        // operation.hashes.push(fullHashId);
        const metadata = Reflect.getMetadata(Metadata_1.REDIS_PROPERTIES, hashClass);
        for (const propMetadata of metadata) {
            const propType = propMetadata.propertyType;
            if (hasPrototypeOf_1.hasPrototypeOf(propType, Set)) {
                const collId = this.getCollectionId(fullHashId, propMetadata);
                if (propMetadata.isRelation) {
                    if (!skipRelations.includes(propMetadata.propertyName)) {
                        operation.relationMappings.push({
                            ownerId: fullHashId,
                            relationClass: propMetadata.relationType,
                            id: collId,
                            type: "set"
                        });
                        operation.sets.push(collId);
                    }
                }
                else {
                    operation.sets.push(collId);
                }
            }
            else if (hasPrototypeOf_1.hasPrototypeOf(propType, Map)) {
                const collId = this.getCollectionId(fullHashId, propMetadata);
                if (propMetadata.isRelation) {
                    if (!skipRelations.includes(propMetadata.propertyName)) {
                        operation.relationMappings.push({
                            ownerId: fullHashId,
                            relationClass: propMetadata.relationType,
                            id: collId,
                            type: "map"
                        });
                        operation.hashes.push(collId);
                    }
                }
                else {
                    operation.hashes.push(collId);
                }
            }
            else if (!hasPrototypeOf_1.hasPrototypeOf(propType, LazySet_1.LazySet) && !hasPrototypeOf_1.hasPrototypeOf(propType, LazyMap_1.LazyMap)) {
                if (propMetadata.isRelation && !skipRelations.includes(propMetadata.propertyName)) {
                    operation.relationMappings.push({
                        ownerId: fullHashId,
                        relationClass: propMetadata.relationType,
                        id: propMetadata.propertyRedisName,
                        type: "key"
                    });
                }
            }
        }
        return operation;
    }
    /**
     * Recursively process hash and set/update metadata
     *
     * @param hashObject
     * @param processedHashes
     */
    updateMetadataInHash(hashObject, processedHashes = []) {
        this.checkMetadata(hashObject);
        if (processedHashes.includes(hashObject)) {
            return;
        }
        processedHashes.push(hashObject);
        const fullHashId = this.getFullIdForEntityObject(hashObject);
        const metadatas = Reflect.getMetadata(Metadata_1.REDIS_PROPERTIES, hashObject.constructor);
        for (const propMetadata of metadatas) {
            if (hasPrototypeOf_1.hasPrototypeOf(propMetadata.propertyType, Set) || hasPrototypeOf_1.hasPrototypeOf(propMetadata.propertyType, Map) ||
                hasPrototypeOf_1.hasPrototypeOf(propMetadata.propertyType, LazySet_1.LazySet) || hasPrototypeOf_1.hasPrototypeOf(propMetadata.propertyType, LazyMap_1.LazyMap)) {
                const collection = hashObject[propMetadata.propertyName];
                if (collection === null) {
                    // Null value for collection
                    Reflect.defineMetadata(Metadata_1.REDIS_VALUE, this.serializeValue(null), hashObject, propMetadata.propertyName);
                }
                else if (collection instanceof LazyMap_1.LazyMap || collection instanceof LazySet_1.LazySet) {
                    // LazySets/Maps don't have REDIS_COLLECTION_VALUE since we don't need to get coll differences for them
                    const collectionId = this.getCollectionId(fullHashId, propMetadata);
                    Reflect.defineMetadata(Metadata_1.REDIS_VALUE, collectionId, hashObject, propMetadata.propertyName);
                }
                else if ((collection instanceof Set || collection instanceof Map) && collection.size > 0) {
                    // Collection have both REDIS_VALUE and REDIS_COLLECTION_VALUE
                    const collectionValueForProp = this.getCollectionId(fullHashId, propMetadata);
                    if (collection instanceof Set) {
                        Reflect.defineMetadata(Metadata_1.REDIS_VALUE, collectionValueForProp, hashObject, propMetadata.propertyName);
                        const setValues = [...collection.values()].map(val => propMetadata.isRelation
                            ? this.prepareRelationValue(val, propMetadata)
                            : this.serializeValue(val)).filter(val => typeof val !== "undefined");
                        Reflect.defineMetadata(Metadata_1.REDIS_COLLECTION_VALUE, setValues, hashObject, propMetadata.propertyName);
                    }
                    else {
                        Reflect.defineMetadata(Metadata_1.REDIS_VALUE, collectionValueForProp, hashObject, propMetadata.propertyName);
                        const mapValues = {};
                        for (const [key, val] of collection) {
                            const preparedKey = this.serializeValue(key);
                            const preparedVal = propMetadata.isRelation
                                ? this.prepareRelationValue(val, propMetadata)
                                : this.serializeValue(val);
                            if (typeof preparedKey !== "undefined" && typeof preparedVal !== "undefined") {
                                mapValues[preparedKey] = preparedVal;
                            }
                        }
                        Reflect.defineMetadata(Metadata_1.REDIS_COLLECTION_VALUE, mapValues, hashObject, propMetadata.propertyName);
                    }
                    // Set metadata in relation hashes if cascase insert/update
                    if (propMetadata.isRelation && (propMetadata.relationOptions.cascadeInsert || propMetadata.relationOptions.cascadeUpdate)) {
                        for (const val of collection.values()) {
                            this.updateMetadataInHash(val, processedHashes);
                        }
                    }
                }
                else {
                    // Delete metadata
                    Reflect.defineMetadata(Metadata_1.REDIS_VALUE, undefined, hashObject, propMetadata.propertyName);
                    Reflect.defineMetadata(Metadata_1.REDIS_COLLECTION_VALUE, undefined, hashObject, propMetadata.propertyName);
                }
            }
            else {
                const value = hashObject[propMetadata.propertyName];
                const preparedValue = propMetadata.isRelation
                    ? this.prepareRelationValue(hashObject[propMetadata.propertyName], propMetadata)
                    : this.serializeValue(hashObject[propMetadata.propertyName]);
                if (typeof preparedValue !== "undefined") {
                    Reflect.defineMetadata(Metadata_1.REDIS_VALUE, preparedValue, hashObject, propMetadata.propertyName);
                    // set metadata for relation
                    if (propMetadata.isRelation && (propMetadata.relationOptions.cascadeInsert || propMetadata.relationOptions.cascadeUpdate) && value) {
                        this.updateMetadataInHash(value, processedHashes);
                    }
                }
                else {
                    Reflect.defineMetadata(Metadata_1.REDIS_VALUE, undefined, hashObject, propMetadata.propertyName);
                }
            }
        }
    }
    /**
     * Recursively process given hash and reset metadata
     *
     * @param entityObj
     */
    resetMetadataInEntityObject(entityObj) {
        this.checkMetadata(entityObj);
        const metadatas = Reflect.getMetadata(Metadata_1.REDIS_PROPERTIES, entityObj.constructor);
        for (const propMetadata of metadatas) {
            // Since we don't support cascade delete no need to delete metadata in relations
            // Reflect.deleteMetadata doesn't work for some reason
            Reflect.defineMetadata(Metadata_1.REDIS_VALUE, undefined, entityObj, propMetadata.propertyName);
            if (hasPrototypeOf_1.hasPrototypeOf(Set, propMetadata.propertyType) || hasPrototypeOf_1.hasPrototypeOf(Map, propMetadata.propertyType)) {
                Reflect.defineMetadata(Metadata_1.REDIS_COLLECTION_VALUE, undefined, entityObj, propMetadata.propertyName);
            }
        }
    }
    /**
     * Hydrate redis data
     *
     * @param hydrationData
     * @returns
     */
    hydrateData(hydrationData) {
        const processedData = new Map();
        const processData = (data) => {
            if (!data || processedData.has(data.id)) {
                return;
            }
            const { entityClass, id, redisData } = data;
            if (!redisData) {
                processedData.set(id, undefined);
                return;
            }
            const dataType = id.slice(0, 2);
            if (dataType === "e:" && !Array.isArray(redisData) && entityClass) {
                const entity = new entityClass();
                processedData.set(id, entity);
                const metadata = Metadata_1.getEntityProperties(entityClass);
                if (metadata) {
                    for (const key of Object.keys(redisData)) {
                        const metadataForKey = metadata.find(m => m.propertyRedisName === key);
                        if (metadataForKey) {
                            const val = redisData[key];
                            const valType = val.slice(0, 2);
                            // define redis value metadata
                            Reflect.defineMetadata(Metadata_1.REDIS_VALUE, val, entity, metadataForKey.propertyName);
                            let valueToSet;
                            if (valType === "e:") {
                                // single relation
                                // process relation if we didn't do it yet
                                if (!processedData.has(val)) {
                                    processData(hydrationData.get(val));
                                }
                                // set relation
                                valueToSet = processedData.get(val);
                            }
                            else if (valType === "a:" || valType === "m:") {
                                // hash or sets
                                if (!processedData.has(val)) {
                                    processData(hydrationData.get(val));
                                }
                                valueToSet = processedData.get(val);
                                const collHydrationData = hydrationData.get(val);
                                if (collHydrationData) {
                                    Reflect.defineMetadata(Metadata_1.REDIS_COLLECTION_VALUE, collHydrationData.redisData, entity, metadataForKey.propertyName);
                                }
                            }
                            else {
                                valueToSet = this.unserializeValue(val);
                            }
                            if (typeof valueToSet !== "undefined") {
                                entity[metadataForKey.propertyName] = valueToSet;
                            }
                        }
                    }
                }
            }
            else if (dataType === "a:" && Array.isArray(redisData) && redisData.length > 0) {
                const set = new Set();
                processedData.set(id, set);
                for (const setData of redisData) {
                    const setDataType = setData.slice(0, 2);
                    // Entity in set
                    if (setDataType === "e:") {
                        // check if we have already processed this entity, otherwise process it
                        if (!processedData.has(setData)) {
                            processData(hydrationData.get(setData));
                        }
                        // Will be processed at this time
                        set.add(processedData.get(setData));
                    }
                    else {
                        set.add(this.unserializeValue(setData));
                    }
                }
            }
            else if (dataType === "m:" && !Array.isArray(redisData)) {
                const map = new Map();
                processedData.set(id, map);
                for (const key of Object.keys(redisData)) {
                    const dataVal = redisData[key];
                    const unserializedKey = this.unserializeValue(key);
                    const type = dataVal.slice(0, 2);
                    if (type === "e:") {
                        if (!processedData.has(dataVal)) {
                            processData(hydrationData.get(dataVal));
                        }
                        map.set(unserializedKey, processedData.get(dataVal));
                    }
                    else {
                        map.set(unserializedKey, this.unserializeValue(dataVal));
                    }
                }
            }
            else {
                processedData.set(id, undefined);
            }
        };
        for (const data of hydrationData.values()) {
            processData(data);
        }
        return [...processedData.values()];
    }
    /**
     * Serialize value for persiting
     *
     * @private
     * @param value
     * @returns
     */
    serializeValue(value) {
        if (typeof value === "number") {
            return `i:${value.toString()}`;
        }
        else if (typeof value === "string") {
            return `s:${value}`;
        }
        else if (typeof value === "boolean") {
            return value ? `b:1` : `b:0`;
        }
        else if (typeof value === "symbol" || typeof value === "function" || typeof value === "undefined") {
            return undefined;
        }
        else if (typeof value === "object") {
            if (value instanceof Date) {
                return `d:${value.getTime().toString()}`;
            }
            else if (value === null) {
                return "null";
            }
            else {
                return `j:${JSON.stringify(value)}`;
            }
        }
        else {
            return `j${JSON.stringify(value)}`;
        }
    }
    /**
     * Unserialize value
     *
     * @private
     * @param value
     * @returns
     */
    unserializeValue(value) {
        if (value === "null") {
            return null;
        }
        const type = value.slice(0, 2);
        const valWithoutType = value.slice(2);
        switch (type) {
            case "i:": return parseFloat(valWithoutType);
            case "s:": return valWithoutType;
            case "b:": return valWithoutType === "1" ? true : false;
            case "d:": return new Date(parseInt(valWithoutType));
            case "j:": return JSON.parse(valWithoutType);
            // return unserialized value for entities, sets and maps
            case "e:": return value;
            case "a:": return value;
            case "m:": return value;
        }
        return undefined;
    }
    /**
     * Get id for collection for property
     *
     * @param entityId
     * @param propertyMetadata
     * @returns
     */
    getCollectionId(entityId, propertyMetadata) {
        if (hasPrototypeOf_1.hasPrototypeOf(propertyMetadata.propertyType, Set) || hasPrototypeOf_1.hasPrototypeOf(propertyMetadata.propertyType, LazySet_1.LazySet)) {
            return `a:${entityId}:${propertyMetadata.propertyRedisName}`;
        }
        else if (hasPrototypeOf_1.hasPrototypeOf(propertyMetadata.propertyType, Map) || hasPrototypeOf_1.hasPrototypeOf(propertyMetadata.propertyType, LazyMap_1.LazyMap)) {
            return `m:${entityId}:${propertyMetadata.propertyRedisName}`;
        }
        else {
            return undefined;
        }
    }
    /**
     * Check metadata correctness
     *
     * @private
     * @param entity
     */
    checkMetadata(entity) {
        const entityType = typeof entity === "object" ? entity.constructor : entity;
        const hashName = Reflect.getMetadata(Metadata_1.REDIS_ENTITY, entityType);
        if (!hashName) {
            throw new Errors_1.MetadataError(entityType, "Class must be decorated with @Entity decorator");
        }
        const metadatas = Reflect.getMetadata(Metadata_1.REDIS_PROPERTIES, entityType);
        if (!metadatas || metadatas.length === 0) {
            throw new Errors_1.MetadataError(entityType, "No any properties");
        }
        const idPropMetadata = metadatas.find(prop => !prop.isRelation && prop.isIdentifyColumn);
        if (!idPropMetadata) {
            throw new Errors_1.MetadataError(entityType, "Must contain one @IdentifyProperty decorator");
        }
    }
    /**
     * Get entity name
     *
     * @private
     * @param entity
     * @returns
     */
    getFullIdForEntityObject(entity) {
        this.checkMetadata(entity);
        const hashId = Metadata_1.getEntityFullId(entity);
        if (typeof hashId === "undefined") {
            throw new Errors_1.MetadataError(entity.constructor, "Unable to to get hash id");
        }
        return hashId;
    }
    /**
     * Get full id for given hash class and id
     *
     * @private
     * @param entityClass
     * @param id
     * @returns
     */
    getFullIdForEntityClass(entityClass, id) {
        const hashId = Metadata_1.getEntityFullId(entityClass, id);
        if (!hashId) {
            throw new Errors_1.MetadataError(entityClass, "Not a redis hash. Perhaps you forgot to add @Hash decorator");
        }
        return hashId;
    }
    /**
     * Make sure that relations are valid and doesn't have same ids but different object links
     * Same id and same object link (i.e rel === rel) is allowed
     *
     * @private
     * @param relations
     * @param parentEntity
     */
    ensureValidRelations(relations, parentEntity) {
        const processedRelations = {};
        for (const rel of relations) {
            const relFullId = this.getFullIdForEntityObject(rel);
            if (processedRelations[relFullId] && processedRelations[relFullId] !== rel) {
                throw new Errors_1.DuplicateIdsInEntityError(parentEntity, relFullId);
            }
            processedRelations[relFullId] = rel;
        }
    }
    /**
     * Prepare relation value persisting. This will be just entity name
     *
     * @private
     * @param value
     * @param metadata
     * @returns
     */
    prepareRelationValue(value, metadata) {
        this.checkMetadata(metadata.relationType);
        if (value) {
            return this.getFullIdForEntityObject(value);
        }
        else if (value === null) {
            return "null";
        }
        else {
            return undefined;
        }
    }
}
exports.Operator = Operator;
//# sourceMappingURL=Operator.js.map