import { Connection } from "../Connection/Connection";
import { EntitySubscriberInterface } from "../Subscriber/EntitySubscriberInterface";
import { Operator } from "./Operator";
export declare type EntityType<T> = {
    new (): T;
} | Function;
/**
 * Main manager to get/save/remove entities
 *
 * @export
 * @class RedisManager
 */
export declare class RedisManager {
    /**
     * Connection instance
     *
     * @protected
     */
    connection: Connection;
    /**
     * Array of entity subscribers
     *
     * @protected
     */
    protected subscribers: Array<EntitySubscriberInterface<any>>;
    /**
     * Entity operator
     *
     * @protected
     */
    protected operator: Operator;
    /**
     * Creates an instance of RedisManager.
     * @param connection
     * @param subscribers
     */
    constructor(connection: Connection, subscribers: Array<EntitySubscriberInterface<any>>);
    /**
     * Save entity
     *
     * @template T
     * @param entity
     * @returns
     */
    save<T extends object>(entity: T): Promise<void>;
    /**
     * Get entity
     *
     * @template T
     * @param entityClass
     * @param id
     * @param [skipRelations]
     * @returns
     */
    load<T>(entityClass: EntityType<T>, id: string | number, skipRelations?: Array<keyof T>): Promise<T | undefined>;
    /**
     * Get entities
     *
     * @template T
     * @param entityClass
     * @param id
     * @param [skipRelations]
     * @returns
     */
    load<T>(entityClass: EntityType<T>, id: string[] | number[], skipRelations?: Array<keyof T>): Promise<T[] | undefined>;
    /**
     * Remove entity. Doesn't remove linked relations
     *
     * @param entity
     * @returns
     */
    remove(entity: object): Promise<void>;
    /**
     * Remove entity be it's id. This WON'T trigger entity subscribers beforeRemove/afterRemove
     *
     * @param entityClass
     * @param id
     * @returns
     */
    removeById(entityClass: EntityType<any>, id: string | number | string[] | number[]): Promise<void>;
    /**
     * Serialize simple value to store it in redis
     *
     * @param value
     * @returns
     */
    serializeSimpleValue(value: any): string | undefined;
    /**
     * Unserialize value
     *
     * @param value
     * @returns
     */
    unserializeSimpleValue(value: string): any;
    /**
     * Init lazy set and maps. This will replace LazySet/LazyMap instances with redis-backed analogs
     *
     * @param entity
     */
    initLazyCollections(entity: {
        [key: string]: any;
    }): void;
    /**
     * Return all entities which should fire related subscribers for save/update/delete operation
     *
     * @private
     * @param entity
     * @param operation
     * @param [entities=[]]
     * @returns
     */
    private getEntitiesForSubscribers(entity, entities?);
    /**
     * Return entities which were somehow changed with given operation
     *
     * @private
     * @param entities
     * @param operation
     * @returns
     */
    private filterEntitiesForPersistenceOperation(entities, operation);
    /**
     * True true if persistence operation is empty (i.e. doesn't have any changes)
     *
     * @private
     * @param operation
     * @returns
     */
    private isEmptyPersistenceOperation(operation);
}
