"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const LazyMap_1 = require("../Collections/LazyMap");
const LazySet_1 = require("../Collections/LazySet");
const RedisLazyMap_1 = require("../Collections/RedisLazyMap");
const RedisLazySet_1 = require("../Collections/RedisLazySet");
const Errors_1 = require("../Errors/Errors");
const Metadata_1 = require("../Metadata/Metadata");
const hasPrototypeOf_1 = require("../utils/hasPrototypeOf");
const Operator_1 = require("./Operator");
/**
 * Main manager to get/save/remove entities
 *
 * @export
 * @class RedisManager
 */
class RedisManager {
    /**
     * Creates an instance of RedisManager.
     * @param connection
     * @param subscribers
     */
    constructor(connection, subscribers) {
        /**
         * Array of entity subscribers
         *
         * @protected
         */
        this.subscribers = [];
        this.connection = connection;
        this.subscribers = subscribers;
        this.operator = new Operator_1.Operator();
    }
    /**
     * Save entity
     *
     * @template T
     * @param entity
     * @returns
     */
    save(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            // Run beforeSave subscribers for all entities regardless will they be finally saved or no
            // started from most deep relation entity to root entity
            const allEntities = this.getEntitiesForSubscribers(entity).reverse();
            for (const ent of allEntities) {
                const subscriber = this.subscribers.find(subscriber => subscriber.listenTo() === ent.constructor);
                if (subscriber && subscriber.beforeSave) {
                    subscriber.beforeSave(ent);
                }
            }
            const operation = yield this.operator.getSaveOperation(entity);
            if (this.isEmptyPersistenceOperation(operation)) {
                return;
            }
            // Call entities subscribers
            // Do operation
            yield this.connection.transaction(executor => {
                for (const deleteSet of operation.deletesSets) {
                    executor.del(deleteSet);
                }
                for (const deleteHash of operation.deleteHashes) {
                    executor.del(deleteHash);
                }
                for (const modifySet of operation.modifySets) {
                    if (modifySet.removeValues.length > 0) {
                        executor.srem(modifySet.setName, modifySet.removeValues);
                    }
                    if (modifySet.addValues.length > 0) {
                        executor.sadd(modifySet.setName, modifySet.addValues);
                    }
                }
                for (const modifyHash of operation.modifyHashes) {
                    if (modifyHash.deleteKeys.length > 0) {
                        executor.hdel(modifyHash.hashId, modifyHash.deleteKeys);
                    }
                    if (Object.keys(modifyHash.changeKeys).length > 0) {
                        executor.hmset(modifyHash.hashId, modifyHash.changeKeys);
                    }
                }
            });
            // update metadata
            this.operator.updateMetadataInHash(entity);
            this.initLazyCollections(entity);
            // Call entities subscribers
            for (const ent of this.filterEntitiesForPersistenceOperation(allEntities, operation)) {
                const subscriber = this.subscribers.find(subscriber => subscriber.listenTo() === ent.constructor);
                if (subscriber && subscriber.afterSave) {
                    subscriber.afterSave(ent);
                }
            }
        });
    }
    load(entityClass, id, skipRelations) {
        return __awaiter(this, void 0, void 0, function* () {
            const idsToLoad = Array.isArray(id) ? id : [id];
            const entityIdToClassMap = new Map();
            const loadedData = new Map();
            const rootLoadOperations = idsToLoad.map(id => this.operator.getLoadOperation(id, entityClass, skipRelations)).filter(op => !!op);
            rootLoadOperations.forEach(op => entityIdToClassMap.set(op.entityId, entityClass));
            const recursiveLoadDataWithRelations = (operations) => __awaiter(this, void 0, void 0, function* () {
                const loadedDataForCall = [];
                // easier to use callbacks instead of processing result
                yield this.connection.batch(executor => {
                    for (const op of operations) {
                        executor.hgetall(op.entityId, (err, result) => loadedDataForCall.push({ id: op.entityId, redisData: result, entityClass: entityIdToClassMap.get(op.entityId) }));
                        for (const hash of op.hashes) {
                            executor.hgetall(hash, (err, result) => loadedDataForCall.push({ id: hash, redisData: result }));
                        }
                        for (const set of op.sets) {
                            executor.smembers(set, (err, result) => loadedDataForCall.push({ id: set, redisData: result }));
                        }
                    }
                });
                loadedDataForCall.forEach(data => loadedData.set(data.id, data));
                const allMappings = [];
                for (const op of operations) {
                    allMappings.push(...op.relationMappings);
                }
                const relationOperations = [];
                for (const mapping of allMappings) {
                    const relationClass = mapping.relationClass;
                    switch (mapping.type) {
                        // single relation in key
                        case "key": {
                            const hashVal = loadedDataForCall.find(data => data.id === mapping.ownerId);
                            if (hashVal && hashVal.redisData && !Array.isArray(hashVal.redisData) && hashVal.redisData[mapping.id]) {
                                const relationVal = hashVal.redisData[mapping.id];
                                if (relationVal && !loadedData.has(relationVal)) {
                                    entityIdToClassMap.set(relationVal, relationClass);
                                    relationOperations.push(this.operator.getLoadOperation(relationVal, relationClass));
                                }
                            }
                            break;
                        }
                        // set of relations
                        case "set": {
                            const set = loadedDataForCall.find(data => data.id === mapping.id);
                            if (set && set.redisData && Array.isArray(set.redisData)) {
                                for (const setVal of set.redisData) {
                                    if (!loadedData.has(setVal)) {
                                        entityIdToClassMap.set(setVal, relationClass);
                                        relationOperations.push(this.operator.getLoadOperation(setVal, relationClass));
                                    }
                                }
                            }
                            break;
                        }
                        // map of relations
                        case "map": {
                            const map = loadedDataForCall.find(data => data.id === mapping.id);
                            if (map && map.redisData && !Array.isArray(map.redisData) && Object.keys(map.redisData).length > 0) {
                                for (const key of Object.keys(map.redisData)) {
                                    const relVal = map.redisData[key];
                                    if (!loadedData.has(relVal)) {
                                        entityIdToClassMap.set(relVal, relationClass);
                                        relationOperations.push(this.operator.getLoadOperation(relVal, relationClass));
                                    }
                                }
                            }
                            break;
                        }
                    }
                }
                if (relationOperations.length > 0) {
                    yield recursiveLoadDataWithRelations(relationOperations.filter(op => !!op));
                }
            });
            yield recursiveLoadDataWithRelations(rootLoadOperations);
            const hydratedData = this.operator.hydrateData(loadedData);
            // Init lazy collections
            hydratedData.forEach(data => {
                if (data && data.constructor && Metadata_1.isRedisEntity(data)) {
                    this.initLazyCollections(data);
                }
            });
            // run subscribers in reverse order
            hydratedData.reduceRight((unusued, data) => {
                if (data && data.constructor) {
                    const subscriber = this.subscribers.find(sub => sub.listenTo() === data.constructor);
                    if (subscriber && subscriber.afterLoad) {
                        subscriber.afterLoad(data);
                    }
                }
            }, {});
            return Array.isArray(id)
                ? hydratedData.filter(data => data && data.constructor === entityClass)
                : hydratedData.find(data => data && data.constructor === entityClass);
        });
    }
    /**
     * Remove entity. Doesn't remove linked relations
     *
     * @param entity
     * @returns
     */
    remove(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const operation = this.operator.getDeleteOperation(entity);
            if (this.isEmptyPersistenceOperation(operation)) {
                return;
            }
            // Since we don't support cascade delete no need to process relations for subscribers
            const subscriber = this.subscribers.find(sub => sub.listenTo() === entity.constructor);
            if (subscriber && subscriber.beforeRemove) {
                subscriber.beforeRemove(entity);
            }
            yield this.connection.transaction(executor => {
                for (const deleteSet of operation.deletesSets) {
                    executor.del(deleteSet);
                }
                for (const deleteHash of operation.deleteHashes) {
                    executor.del(deleteHash);
                }
            });
            this.operator.resetMetadataInEntityObject(entity);
            if (subscriber && subscriber.afterRemove) {
                subscriber.afterRemove(entity);
            }
        });
    }
    /**
     * Remove entity be it's id. This WON'T trigger entity subscribers beforeRemove/afterRemove
     *
     * @param entityClass
     * @param id
     * @returns
     */
    removeById(entityClass, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const idsToRemove = Array.isArray(id) ? id : [id];
            const operations = idsToRemove.map(id => this.operator.getDeleteOperation(entityClass, id));
            if (operations.every(this.isEmptyPersistenceOperation)) {
                return;
            }
            yield this.connection.transaction(executor => {
                for (const operation of operations) {
                    for (const deleteSet of operation.deletesSets) {
                        executor.del(deleteSet);
                    }
                    for (const deleteHash of operation.deleteHashes) {
                        executor.del(deleteHash);
                    }
                }
            });
        });
    }
    /**
     * Serialize simple value to store it in redis
     *
     * @param value
     * @returns
     */
    serializeSimpleValue(value) {
        return this.operator.serializeValue(value);
    }
    /**
     * Unserialize value
     *
     * @param value
     * @returns
     */
    unserializeSimpleValue(value) {
        return this.operator.unserializeValue(value);
    }
    /**
     * Init lazy set and maps. This will replace LazySet/LazyMap instances with redis-backed analogs
     *
     * @param entity
     */
    initLazyCollections(entity) {
        const id = Metadata_1.getEntityFullId(entity);
        if (!id) {
            throw new Errors_1.MetadataError(entity.constructor, "Unable to get entity id");
        }
        const properties = Metadata_1.getEntityProperties(entity);
        if (!properties) {
            throw new Errors_1.MetadataError(entity.constructor, "No any properties");
        }
        for (const prop of properties) {
            if (hasPrototypeOf_1.hasPrototypeOf(prop.propertyType, LazyMap_1.LazyMap) && !(entity[prop.propertyName] instanceof RedisLazyMap_1.RedisLazyMap)) {
                const mapId = this.operator.getCollectionId(id, prop);
                if (mapId) {
                    entity[prop.propertyName] = new RedisLazyMap_1.RedisLazyMap(mapId, this, prop.isRelation
                        ? prop.relationType
                        : undefined, prop.isRelation
                        ? prop.relationOptions.cascadeInsert
                        : false);
                }
            }
            else if (hasPrototypeOf_1.hasPrototypeOf(prop.propertyType, LazySet_1.LazySet) && !(entity[prop.propertyName] instanceof RedisLazySet_1.RedisLazySet)) {
                const setId = this.operator.getCollectionId(id, prop);
                if (setId) {
                    entity[prop.propertyName] = new RedisLazySet_1.RedisLazySet(setId, this, prop.isRelation
                        ? prop.relationType
                        : undefined, prop.isRelation
                        ? prop.relationOptions.cascadeInsert
                        : false);
                }
            }
        }
    }
    /**
     * Return all entities which should fire related subscribers for save/update/delete operation
     *
     * @private
     * @param entity
     * @param operation
     * @param [entities=[]]
     * @returns
     */
    getEntitiesForSubscribers(entity, entities = []) {
        if (!Metadata_1.isRedisEntity(entity) || entities.includes(entity)) {
            return entities;
        }
        entities.push(entity);
        const metadata = Metadata_1.getEntityProperties(entity);
        if (!metadata) {
            return entities;
        }
        // const redisHash = getRedisHashId(entity);
        for (const propMetadata of metadata) {
            // no need to process non relations or without cascading options
            if (!propMetadata.isRelation || !(propMetadata.relationOptions.cascadeInsert || propMetadata.relationOptions.cascadeUpdate)) {
                continue;
            }
            const propValue = entity[propMetadata.propertyName];
            if (!propValue) {
                continue;
            }
            if (propValue instanceof Map || propValue instanceof Set) {
                for (const collVal of propValue.values()) {
                    this.getEntitiesForSubscribers(collVal, entities);
                }
            }
            else {
                this.getEntitiesForSubscribers(propValue, entities);
            }
        }
        return entities;
    }
    /**
     * Return entities which were somehow changed with given operation
     *
     * @private
     * @param entities
     * @param operation
     * @returns
     */
    filterEntitiesForPersistenceOperation(entities, operation) {
        return entities.filter(entity => {
            const hashId = Metadata_1.getEntityFullId(entity);
            if (!hashId) {
                return false;
            }
            if (operation.deleteHashes.find(name => name.includes(hashId))) {
                return true;
            }
            if (operation.deletesSets.find(name => name.includes(hashId))) {
                return true;
            }
            if (operation.modifyHashes.find(v => ((v.deleteKeys.length > 0 || Object.keys(v.changeKeys).length > 0) && v.hashId.includes(hashId)))) {
                return true;
            }
            if (operation.modifySets.find(v => ((v.addValues.length > 0 || v.removeValues.length > 0) && v.setName.includes(hashId)))) {
                return true;
            }
            return false;
        });
    }
    /**
     * True true if persistence operation is empty (i.e. doesn't have any changes)
     *
     * @private
     * @param operation
     * @returns
     */
    isEmptyPersistenceOperation(operation) {
        return operation.deleteHashes.length === 0 && operation.deletesSets.length === 0 &&
            (operation.modifyHashes.length === 0 || operation.modifyHashes.every(val => val.deleteKeys.length === 0 && Object.keys(val.changeKeys).length === 0)) &&
            (operation.modifySets.length === 0 || operation.modifySets.every(val => val.addValues.length === 0 && val.removeValues.length === 0));
    }
}
exports.RedisManager = RedisManager;
//# sourceMappingURL=RedisManager.js.map