"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Connection_1 = require("../Connection/Connection");
const redis = require("../utils/PromisedRedis");
/**
 * Get available database from pool of databases (presented as SET in redis)
 *
 * @returns
 */
function getAvailableDatabase() {
    return __awaiter(this, void 0, void 0, function* () {
        const client = createRawRedisClient();
        let dbNumber = yield client.spopAsync("available_db");
        // Set doesn't exist
        if (!dbNumber) {
            // Create set of db connetions
            yield client.saddAsync("available_db", ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13", "14", "15"]);
            // Obtain again
            dbNumber = yield client.spopAsync("available_db");
        }
        if (!dbNumber) {
            throw new Error("Unable to obtain database number");
        }
        yield client.quitAsync();
        return dbNumber;
    });
}
exports.getAvailableDatabase = getAvailableDatabase;
/**
 * Release available database and return it to the pool
 *
 * @export
 * @param dbNumber
 * @returns
 */
function releaseDatabase(dbNumber) {
    return __awaiter(this, void 0, void 0, function* () {
        const client = redis.createClient(parseInt(process.env.REDIS_PORT) || 6379, process.env.REDIS_HOST);
        yield client.saddAsync("available_db", dbNumber);
        yield client.quitAsync();
    });
}
exports.releaseDatabase = releaseDatabase;
/**
 * Create redis client
 *
 * @export
 * @returns
 */
function createRawRedisClient() {
    const client = redis.createClient(parseInt(process.env.REDIS_PORT) || 6379, process.env.REDIS_HOST);
    return client;
}
exports.createRawRedisClient = createRawRedisClient;
// Just in case if the global state will be shared between tests
const connectionToDbNumber = new WeakMap();
/**
 * Create redis connection
 *
 * @param [options={}]
 * @export
 * @returns
 */
function createRedisConnection(options = {}) {
    return __awaiter(this, void 0, void 0, function* () {
        const dbNumber = yield getAvailableDatabase();
        const connection = new Connection_1.Connection();
        connectionToDbNumber.set(connection, dbNumber);
        try {
            yield connection.connect(Object.assign({ db: dbNumber, host: process.env.REDIS_HOST, port: parseInt(process.env.REDIS_PORT) || 6379 }, options));
        }
        catch (e) {
            yield releaseDatabase(dbNumber);
            throw e;
        }
        return connection;
    });
}
exports.createRedisConnection = createRedisConnection;
/**
 * Clean & release redis connection
 *
 * @export
 * @param connection
 * @returns
 */
function cleanRedisConnection(connection) {
    return __awaiter(this, void 0, void 0, function* () {
        if (connection.isConnected) {
            yield connection.flushdb();
        }
        const dbNumber = connectionToDbNumber.get(connection);
        if (dbNumber) {
            yield releaseDatabase(dbNumber);
        }
        if (connection.isConnected) {
            yield connection.disconnect();
        }
    });
}
exports.cleanRedisConnection = cleanRedisConnection;
/**
 * Return database number used for connection
 *
 * @export
 * @param connection
 * @returns
 */
function getDatabaseNumberForConnection(connection) {
    return connectionToDbNumber.get(connection);
}
exports.getDatabaseNumberForConnection = getDatabaseNumberForConnection;
//# sourceMappingURL=redis.js.map